# Claudio Perez
import re, os
import fnmatch
import logging
from pathlib import Path
from itertools import product

from . import jsonpointer

from aurore.utils.treeutils import iterate_leaves

logger = logging.getLogger("aurore.selectors")


class Pattern:
    def __init__(self, pattern):
        self.pattern = pattern

    def validate(self,field)->bool:
        if isinstance(field,(str,float,int)):
            logger.debug(f"{self.pattern}, {field}")
            return fnmatch.fnmatch(field,self.pattern)
        else:
            return self.pattern == field


class PathBuilder:
    """
    Examples
    --------
    >>> item = {'id': 'xyz-123', 'a': {'b': 'c'}}
    >>> PathBuilder('%i/inter/file.rst').resolve(item)
    Path('xyz-123/inter/file.rst')
    >>> PathBuilder('%%:id/inter/%%:a:b/file.rst').resolve(item)
    Path('xyz-123/inter/c/file.rst')
    """
    def __init__(self,
        template: str,
        delimeter: str = "/",
        pointer_delimeter:str = ":",
        unpack_fields:bool = False
    ):
        self.template = template.split(delimeter)
        self.pointer_delimeter = pointer_delimeter
        # self.pointers = [
        #     Pointer(k) for k in self.template if "%" in k
        # ]

    def resolve(self,item)->str:
        return os.path.sep.join([
            Pointer(s,delimeter=self.pointer_delimeter).resolve(item)
                if "%" in s else s for s in self.template
        ])

    def resolve_recursively(self,item):
        components = [
            [*Pointer(s,delimeter=self.pointer_delimeter).resolve_recursively(item)]
              if "%" in s else [s] for s in self.template
        ]
        for combo in product(*components):
            yield os.path.sep.join(combo)


class Pointer:
    token_keys = {
        "c": lambda i,_: i["categories"],
        "t": lambda i,_: i["title"],
        "i": lambda i,_: i["id"],
        ".": lambda i,_: i,
        "%": lambda i,_: i,
    }
    attrib_tokens = {
        "s": lambda tree: tree.attrib["src"],
        "b": lambda tree: tree.attrib["base"]
    }

    def __init__(self, pointer:str,
        recurse=False,maxlen:int=30,truncate:bool=False,
        bracket_as_slice=False, delimeter="/"
    ):
        self.recurse:  bool = recurse
        self.truncate: bool = truncate
        self.slice:   slice = slice(maxlen)
        self.delimeter = delimeter
        if bracket_as_slice:
            match = re.search(r"\[(.*)\]$",pointer)
            if match:
                slice_str = match.group(1)
                self.slice = slice(*[
                    {True: lambda n: None, False: int}[x == ''](x)
                        for x in (slice_str.split(':') + ['', '', ''])[:3]
                ])
                pointer = pointer.rsplit("[",1)[0]
                logger.debug(f"slice_str: {slice_str}")
                logger.debug(f"slice: {self.slice}")

        if pointer[0] == "%":
            self.tokens = re.split(
                "([%|{0}][^{0}]*)".format("\\" + delimeter if delimeter in ["."] else delimeter
            ), pointer)[1::2]

            logger.debug(f"tokens: {self.tokens}")
        else:
            raise Exception(
                f"Unimplemented pattern handler for pointer: "
                f"{pointer}"
                )

    def resolve_tokens(self,item):
        for token in self.tokens:
            if re.match("^%.*", token):
                item = Pointer.token_keys[token[1]](item,None)
            else:
            # elif re.match("^/.*",token):
                item = jsonpointer.resolve_pointer(item, token, delim=self.delimeter)
        return item

    def resolve(self,item):
        base = self.resolve_tokens(item)
        return base[self.slice] \
            if self.truncate and isinstance(base,str) \
            else base

    def resolve_recursively(self,item):
        base = self.resolve_tokens(item)
        for i in iterate_leaves(base):
            yield i

class TrueSelector:
    def __init__(self, *args, **kwds):
        pass 
    def validate(self, item)->bool:
        return True

class FalseSelector:
    def __init__(self, *args, **kwds):
        pass 
    def validate(self, item)->bool:
        return False

class Selector:
    def __init__(self,selector:str):

        if ":" in selector:
            recurse = True
            pointer, pattern = selector.split(":")
        elif "=" in selector:
            recurse = False
            pointer, pattern = selector.split("=")
        else:
            recurse = False
            pointer, pattern = r"%i", selector

        self.pattern = Pattern(pattern)
        self.pointer = Pointer(pointer,recurse=recurse)
        self.recurse = recurse


    def validate(self,item):
        if self.recurse:
            return any(
                self.pattern.validate(val)
                    for val in self.pointer.resolve_recursively(item)
            )
        else:
            return self.pattern.validate(self.pointer.resolve(item))


def check_includes(args, item):
    logger.debug(args.include_item)
    if args.include_item:
        inclusive_selectors = [Selector(i) for i in args.include_item]
    else:
        inclusive_selectors = [TrueSelector()]
    
    if args.include_exclusive:
        exclusive_selectors = [Selector(i) for i in args.include_exclusive]
    else:
        exclusive_selectors = [TrueSelector()]
    
    if args.exclude_item:
        exclude_selectors = [Selector(i) for i in args.exclude_item]
    else:
        exclude_selectors = [FalseSelector()]

    
    return any(selector.validate(item) for selector in inclusive_selectors) \
           and \
           all(selector.validate(item) for selector in exclusive_selectors) \
           and \
           all(not selector.validate(item) for selector in exclude_selectors)





__version__ = "0.0.8"

def load(*args,include='ccorp',include_dirs=None,**kwds):
    if include == 'ccorp':
        from .ccorp_include import YAML
        yaml = YAML(typ='safe', pure=True)
        yaml.allow_duplicate_keys = True
        yaml.include_dirs = include_dirs
        return yaml.load(*args, **kwds)

    elif include=='ruamel':
        from .ruamel_include import YAML, yaml_include, my_compose_document
        yaml = YAML(typ='safe', pure=True)
        yaml.Composer.compose_document = my_compose_document
        yaml.default_flow_style = False
        yaml.Constructor.add_constructor("!include", yaml_include)

        return yaml.load(*args, **kwds)






import os
import sys
import logging
from collections import UserDict

from aurore import exceptions
from aurore import utils

from yaml import YAMLError

log = logging.getLogger('aurore.config')


class ValidationError(Exception):
    """Raised during the validation process of the config on errors."""


class Config(UserDict):
    """
    aurore Configuration dict

    This is a fairly simple extension of a standard dictionary. It adds methods
    for running validation on the structure and contents.
    """

    def __init__(self, schema,config_file_path=None):
        """
        The schema is a Python dict which maps the config name to a validator.
        """

        self._schema = schema
        self._schema_keys = set(dict(schema).keys())
        # Ensure config_file_path is a Unicode string
        if config_file_path is not None and not isinstance(config_file_path, str):
            try:
                # Assume config_file_path is encoded with the file system encoding.
                config_file_path = config_file_path.decode(encoding=sys.getfilesystemencoding())
            except UnicodeDecodeError:
                raise ValidationError("config_file_path is not a Unicode string.")
        self.config_file_path = config_file_path
        self.data = {}

        self.user_configs = []
        self.set_defaults()

    def set_defaults(self):
        """
        Set the base config by going through each validator and getting the
        default if it has one.
        """

        for key, config_option in self._schema:
            self[key] = config_option.default

    def _validate(self):

        failed, warnings = [], []

        for key, config_option in self._schema:
            try:
                value = self.get(key)
                self[key] = config_option.validate(value)
                warnings.extend([(key, w) for w in config_option.warnings])
                config_option.reset_warnings()
            except ValidationError as e:
                failed.append((key, e))

        for key in (set(self.keys()) - self._schema_keys):
            warnings.append((
                key, "Unrecognised configuration name: {}".format(key)
            ))

        return failed, warnings

    def _pre_validate(self):

        failed, warnings = [], []

        for key, config_option in self._schema:
            try:
                config_option.pre_validation(self, key_name=key)
                warnings.extend([(key, w) for w in config_option.warnings])
                config_option.reset_warnings()
            except ValidationError as e:
                failed.append((key, e))

        return failed, warnings

    def _post_validate(self):

        failed, warnings = [], []

        for key, config_option in self._schema:
            try:
                config_option.post_validation(self, key_name=key)
                warnings.extend([(key, w) for w in config_option.warnings])
                config_option.reset_warnings()
            except ValidationError as e:
                failed.append((key, e))

        return failed, warnings

    def validate(self):

        failed, warnings = self._pre_validate()

        run_failed, run_warnings = self._validate()

        failed.extend(run_failed)
        warnings.extend(run_warnings)

        # Only run the post validation steps if there are no failures, warnings
        # are okay.
        if len(failed) == 0:
            post_failed, post_warnings = self._post_validate()
            failed.extend(post_failed)
            warnings.extend(post_warnings)

        return failed, warnings

    def load_dict(self, patch):

        if not isinstance(patch, dict):
            raise exceptions.ConfigurationError(
                "The configuration is invalid. The expected type was a key "
                "value mapping (a python dict) but we got an object of type: "
                "{}".format(type(patch)))

        self.user_configs.append(patch)
        self.data.update(patch)

    def load_file(self, config_file):
        try:
            return self.load_dict(utils.yaml_load(config_file))
        except YAMLError as e:
            # aurore knows and understands ConfigurationErrors
            raise exceptions.ConfigurationError(
                "aurore encountered as error parsing the configuration file: {}".format(e)
            )


def _open_config_file(config_file):

    # Default to the standard config filename.
    if config_file is None:
        config_file = os.path.abspath('.aurore/config.yaml')

    # If closed file descriptor, get file path to reopen later.
    if hasattr(config_file, 'closed') and config_file.closed:
        config_file = config_file.name

    log.debug("Loading configuration file: {}".format(config_file))

    # If it is a string, we can assume it is a path and attempt to open it.
    if isinstance(config_file, str):
        if os.path.exists(config_file):
            config_file = open(config_file, 'rb')
        else:
            raise exceptions.ConfigurationError(
                "Config file '{}' does not exist.".format(config_file))

    # Ensure file descriptor is at begining
    config_file.seek(0)

    return config_file


def load_config(config_file=None, **kwargs):
    """
    Load the configuration for a given file object or name

    The config_file can either be a file object, string or None. If it is None
    the default `config.yaml` filename will loaded.

    Extra kwargs are passed to the configuration to replace any default values
    unless they themselves are None.
    """
    options = kwargs.copy()

    # Filter None values from the options. This usually happens with optional
    # parameters from Click.
    for key, value in options.copy().items():
        if value is None:
            options.pop(key)

    config_file = _open_config_file(config_file)
    options['config_file_path'] = getattr(config_file, 'name', '')
    
    # Initialise the config with the default schema .
    from aurore import config
    cfg = Config(schema=config.DEFAULT_SCHEMA, config_file_path=options['config_file_path'])
    # First load the config file
    cfg.load_file(config_file)
    # Then load the options to overwrite anything in the config.
    cfg.load_dict(options)

    errors, warnings = cfg.validate()

    for config_name, warning in warnings:
        log.warning("Config value: '%s'. Warning: %s", config_name, warning)

    for config_name, error in errors:
        log.error("Config value: '%s'. Error: %s", config_name, error)

    for key, value in cfg.items():
        log.debug("Config value: '%s' = %r", key, value)

    if len(errors) > 0:
        pass
        # raise exceptions.ConfigurationError(
        #     "Aborted with {} Configuration Errors!".format(len(errors))
        # )
    elif cfg['strict'] and len(warnings) > 0:
        raise exceptions.ConfigurationError(
            "Aborted with {} Configuration Warnings in 'strict' mode!".format(len(warnings))
        )

    return cfg



import os
import subprocess
import logging
import xml.etree.ElementTree as ElementTree
from datetime import datetime

from .git_utils import sha1
from .uri_utils import resolve_uri
from .proc_rst import find_dependencies
from .utils import norm_join

logger = logging.getLogger("aurore.proc_xml")

def xml_to_map(xml_item: ElementTree, base=None, context=None, relpath=None,**kwds)->dict:
    return proc_elem(xml_item, base, context, relpath, **kwds)[0]


def proc_elem(elem, base, context, relpath=None,**kwds):
    if test_pull_content(elem):
        logger.info(f"Pulling content for src: {elem.attrib['src']}")
        content, sub_state = pull_content(elem, base)
        val, _ = proc[elem.tag](content, base, context=context, relpath=relpath, **kwds)
    else:
        content = elem
        val, sub_state = proc[elem.tag](content, base, context=context, relpath=relpath,**kwds)
    return val, sub_state


def test_pull_content(elem: ElementTree)->bool:
    """test whether content must be pulled for an element."""
    # return not (elem or elem.text) and "src" in elem.attrib
    return "src" in elem.attrib


def pull_content(elem, base, verbose=True)->ElementTree:
    if "base" in elem.attrib:
        base = elem.attrib["base"]

    src = elem.attrib["src"]
    if "proc" in elem.attrib:
        if "type" in elem.attrib:
            raise Exception("Unimplemented")
        else:
            content, state = proc_src[elem.attrib["proc"]](src, base)
    else:
        logger.info(f"Resolving {src}, {base}")
        content = resolve_uri(src, base)
        state = None
    try:
        elem.text = content.text
    except:
        elem.text = content
    logger.info(f"Pulled content: {content}")
    return elem, {"src": state}


def proc_map(elem, base, context=None, **kwds):
    if "base" in elem.attrib:
        base = elem.attrib["base"]
    logger.info(f"key: {elem.attrib['key'] if 'key' in elem.attrib else ''}, base: {base}")
    output = {}
    state = {}
    for el in elem:
        key = el.attrib["key"] if "key" in el.attrib else el.attrib["id"]
        val, sub_state = proc_elem(el, base, context, **kwds)
        output.update({key : val})
        if sub_state:
            state.update({key: sub_state})
    return output, state

def proc_set(elem, base, context=None, **kwds):
    if "base" in elem.attrib:
        base = elem.attrib["base"]
    if "cast" in elem.attrib:
        cast_type = post_proc[elem.attrib["cast"]]
    else:
        cast_type = lambda x, **kwds: x
    output = set()
    state  = set()
    if isinstance(elem.text,(list,set)):
        output.update(elem.text if isinstance(elem.text,set) else set(elem.text))
    for el in elem:
        val, sub_state = proc_elem(el, base, context=context, cast_type=cast_type, **kwds)

        if "unpack" in el.attrib:
            #getattr(output,el.attrib["unpack"])(val)
            unpacker = getattr(output,el.attrib["unpack"])
            unpacker(cast_type(x,base,**kwds) for x in val)
            getattr(state,el.attrib["unpack"])(sub_state)
        else:
            unpacker = output.update
            if isinstance(val,set):
                unpacker(val)
            elif isinstance(val,list):
                unpacker(set(val))
            else:
                output.add(val)
            state.add(sub_state)
        # if el.tail:
        #     output.append(el.tail)
    if not any(state):
        state = []
    return output, state

def proc_list(elem, base, context=None, **kwds):
    if "base" in elem.attrib:
        base = elem.attrib["base"]
    output = []
    state = []
    if isinstance(elem.text,list):
        output.extend(elem.text)
    for el in elem:
        val, sub_state = proc_elem(el, base, context, **kwds)
        logger.info(f" VAL: {val.text if hasattr(val,'text') else val}")
        if "unpack" in el.attrib:
            getattr(output,el.attrib["unpack"])(val)
            getattr(state,el.attrib["unpack"])(sub_state)
        else:
            output.append(val)
            state.append(sub_state)
    if not any(state):
        state = []
    return output, state


def cli_usage(src,base_uri):
    # build_loc = os.path.join(base_uri,src)
    build_loc = norm_join(base_uri, src)
    state = {"sha1": sha1(src)}
    if build_loc[-3:] == ".py":
        try:
            text = subprocess.check_output([
                "python", os.path.expandvars(build_loc), "--help"
            ]).decode().replace("<","&lt")
        except:
            text = ""
    else:
        try:
            text = subprocess.check_output([
                os.path.expandvars(build_loc)
            ]).decode().replace("<","&lt")
        except:
            try:
                text = subprocess.check_output([
                    os.path.expandvars(build_loc), "--help"
                ]).decode().replace("<","&lt")
            except:
                text = ""

    return text, state

def proc_date(elem, base, context=None, **kwds):
    if "-" not in elem.text:
        if len(elem.text)==8:
            date = f"{date[:4]}-{date[4:6]}-{date[6:]}"
        else:
            logger.error(f"Unrecognized date: {elem.text}")
            return elem.text, None
    date = datetime.fromisoformat(elem.text)
    return date.astimezone().isoformat(), None

def proc_var(elem, base, context=None, **kwds):
    if context is None:
        context = {}
    ans = None
    for var in elem:
        ans, state = proc[var.tag](var,base,context=context,**kwds)
        logger.debug(f"var ans: {ans}")
        if "id" in var.attrib:
            context.update({var.attrib["id"]: ans})
        else:
            context.update({"ans": ans})

    if ans is None:
        return elem.text, None
    if "cast" in elem.attrib:
        logger.debug(f"ans precast: {ans}")
        ans = post_proc[elem.attrib["cast"]](ans)
        logger.debug(f"ans post-cast: {ans}")
    return ans, state

def proc_eval(elem, base, context, **kwds):
    context.update({
        "os": os,
        "subprocess": subprocess
    })
    if "base" in elem.attrib:
        base = elem.attrib["base"]
    return eval(elem.text,context), {"sha1": sha1(elem.text)}

def proc_path(el,base,context,relpath,**kwds):
    path = post_path(el.text, base, relpath)
    return path, None

def post_path(path, base=None, relpath=None):
    path = norm_join(base,path)
    if relpath:
        logger.debug(f"Normalizing path relative to: {relpath}")
        path = os.path.relpath(os.path.normpath(path), os.path.abspath(relpath))
    return path

# def proc_str(el, base, context, **kwds):
#     try:
#         return str(el.text.text)
#     except:
#         return str(el.text)

proc = {
    "var": proc_var,
    "str": lambda el,_,**__: (str(el.text), None),
    "int": lambda el,_,**__: (int(el.text), None),
    "flt": lambda el,_,**__: (float(el.text), None),
    "uri": lambda el,_,**__: (str(el.text), None),
    "lst": proc_list,
    "set": proc_set,
    "tag": lambda el,_,**__: (bool(el.text),None),
    "map": proc_map,
    "path": proc_path,
    "date": proc_date,
    "item": lambda el,_,**__: (el.text, None),
    "bool": lambda el,_,**__: (bool(el.text), None),
    "eval": proc_eval,
    "date": lambda el,_,**__: (str(el.text),None),
}

post_proc = {
    "lst": list,
    "map": dict,
    "int": int,
    "str": str,
    "path": post_path,
}

proc_src = {
    "aurore.cli_usage": cli_usage,
    "aurore.find_dependencies": find_dependencies
}



#Claudio Perez
import os
import re
import docutils.utils
from docutils.core import publish_doctree
import xml.etree.ElementTree as ElementTree
from .utils import norm_join

report_level = docutils.utils.Reporter.SEVERE_LEVEL + 1

def rst_to_xml(filename):
    with open(filename,"r") as f:
        doctree = publish_doctree(
                f.read(),
                settings_overrides={'report_level':report_level}
                )
    xml_representation = doctree.asdom().toxml()
    return ElementTree.fromstring(xml_representation)

def find_dependencies(src: str, base:str)->list:
    rst_tree = rst_to_xml(
        os.path.expandvars(
            norm_join(base,src)
        )
    )
    dependencies = []
    lit_includes = [
        i.text for i in rst_tree.findall(".//literal_block")
            if i.text and "literalinclude" in i.text
    ]
    paths = [
        i.split("\n")[0].split("literalinclude:: ")[1] for i in lit_includes
    ]
    dependencies.extend(paths)
    dependencies.extend([
        image.attrib["uri"] for image in rst_tree.findall(".//image")
        ])

    return dependencies, None


def parse_rst(text:str):
    doctree = publish_doctree(
            text,
            settings_overrides={'report_level':report_level}
            ).asdom()

    # Convert to etree.ElementTree since this is easier to work with than
    # xml.minidom
    doctree = ElementTree.fromstring(doctree.toxml())

    # Get all field lists in the document.
    field_lists = doctree.findall('field_list')

    fields = [f for field_list in field_lists \
        for f in field_list.findall('field')]

    field_names = [name.text for field in fields \
        for name in field.findall('field_name')]

    field_text = [ElementTree.tostring(element) for field in fields \
        for element in field.findall('field_body')]

    return dict(zip(field_names, field_text))




import os
import json
import logging
import subprocess
import xml.etree.ElementTree as ElementTree
import xml.dom.minidom as minidom
from datetime import datetime, timezone

from .uri_utils import resolve_uri
from .proc_xml import xml_to_map, proc_var, proc_elem, post_proc

logger = logging.getLogger("aurore.api")

class SetEncoder(json.JSONEncoder):
    def default(self, obj):
       if isinstance(obj, set):
          return list(obj)
       return json.JSONEncoder.default(self, obj)


def feed_init(args, config)->dict:
    pass

#--------------------------------------------
# Post
#--------------------------------------------
def post_init(args, config)->dict:
    if "type" in args: args.base = args.type.split("/")[0]
    #print(args.base)
    lids = [k for k in config.namespace].sort()
    if lids:
        _, num = lids[-1].split("-")
        lid = f"{(int(num) + 1):04}"
    else:
        lid = f"{args.base}-0001"
    with open(config.init_type[args.type], "r") as f:
        lines = f.readlines()

    for i, ln in enumerate(lines):
        lines[i] = ln.replace("$$lid",lid)
        lines[i] = ln.replace("$$published",datetime.now(timezone.utc).astimezone().isoformat())
    accum =  {
        "lid": lid,
        "base": ElementTree.fromstringlist(lines),
        "lines": [line for line in lines if "<input" not in line]
    }
    return accum

def post_close(args, config, accum):
    data = {}
    for i, ipt in enumerate(accum["base"].findall("input")):
        val = input(ipt.attrib["key"]+": ")
        try:
            val = eval(val)
        except:
            pass
        data[ipt.attrib["key"]] = val

    if "$$lid" in args.location:
        destination = args.location.replace("$$lid",accum["lid"])
        os.mkdir(accum["lid"])
    else:
        destination = args.location
    with open(destination,"w+") as f:
        f.writelines(build_template(accum["lines"], data))

def build_template(template:list, data:dict):
    for k,v in data.items():
        for j,ln in enumerate(template):
            template[j] = ln.replace(k, str(v))
    return template

#--------------------------------------------
# Get
#--------------------------------------------

def get_init(args,config)->dict:

    accum = {"items": {}, "categories": {}, "tags": {}}
    accum["categories"] = {
        scheme.attrib["key"]: {
            "map": xml_to_map(scheme.find("map")),
            "var": scheme.find("var")
        } for scheme in args.category_schemes
    }
    accum["tags"] = _tag(args.tag_schemes) if args.tag_schemes else {}
    #print(accum["tags"])
    return accum

def get_item(rsrc, args:object, config:object, accum:dict)->dict:
    if "rel_path" in args:
        if args.rel_path:
            relpath = os.path.expandvars(args.rel_path)
        else:
            # relpath=os.getcwd()
            relpath = None
    else:
        relpath = None
    logger.debug(f"Relative path: {relpath}")

    raw_item: dict = xml_to_map(rsrc, args.base_uri, relpath=relpath)
    logger.debug(raw_item)

    if "base" in rsrc.attrib:
        base = rsrc.attrib["base"]
    else:
        base = args.base_uri

    item = categorize(accum["categories"], raw_item, base) if accum["categories"] else raw_item
    item = tag_item(accum["tags"], item, base) if accum["tags"] else item
    accum['items'].update({rsrc.attrib["id"]: item})
    return accum

def get_close(args, config, accum):
    output = {
        "items":  accum["items"]
    }
    if accum["categories"]:
        output.update({
            "categories": {
                category: accum["categories"][category]["map"] for category in accum["categories"]
            }
        })

    print(json.dumps(output,indent=2,cls=SetEncoder))

def tag_item(tags, map_item: dict, base_uri):
    logger.debug(map_item)
    map_item["tags"] = []
    for tag in tags:
        logger.info(f"Tag base uri: {base_uri}")
        val = proc_elem(tags[tag], base_uri, {"item": map_item})[0]
        logger.info(f"Tag value: {val}")
        if val:
            map_item["tags"].append(tag)

    return map_item

def categorize(categories, map_item: dict, base_uri):
    logger.debug(map_item)
    map_item["categories"] = {}

    for category in categories:
        logger.info(f"Category base uri: {base_uri}")
        val = proc_elem(categories[category]["var"], base_uri, {"item": map_item})[0]
        if val is None: val="None"
        if "cast" in categories[category]["var"].attrib:
            val = post_proc[categories[category]["var"].attrib["cast"]](val)
        logger.info(f"Category value: {val}")
        assert val in categories[category]["map"], f"Key {val} not in mapping {categories[category]['map']}"
        map_item["categories"].update({category: val})

    return map_item

def _tag(tag_schemes:ElementTree, base:str="")->dict:
    #print(f"_tag({tag_schemes.attrib['key']})")
    a = {
            ".".join([base, tag_schemes.attrib["key"], scheme.attrib["key"]]).strip(".") : scheme
                for scheme in tag_schemes.findall("./tag")
    }
    for b in tag_schemes.findall(".//tag-scheme"):
        if "src" in b.attrib:
            #print(b.attrib["src"])
            scheme = resolve_uri(b.attrib["src"])
            #print(scheme)
        else:
            scheme = b

        a.update(_tag(scheme,base=".".join([base,tag_schemes.attrib["key"]])))
    #print(f"a: {a}")
    return a




import os
import json
import logging
from urllib.parse import urlparse
import xml.etree.ElementTree as ElementTree

import yaml

import aurore.jsonpointer
from aurore.utils import norm_join

# from aurore.fpointer import resolve_fpointer
from . import proc_rst

logger = logging.getLogger("aurore.uri_utils")

def extract_uri(
    scheme="aurore",
    base="",
    path=""
    ): pass


def resolve_uri(ref:str, local_base="",fmt=None)->ElementTree:
    ref = norm_join(local_base, ref)
    # if local_base:
    #     ref = os.path.join(local_base, ref)
    # ref = os.path.expandvars(ref)
    url = urlparse(ref)

    if os.path.isfile(url.path):
        _, ext = os.path.splitext(url.path)
        fmt = fmt if fmt else ext
        if fmt in [".xml", ".html"]:
            item = ElementTree.parse(url.path)
            if url.fragment:
                if url.fragment[0] not in ['.', '/']:
                    return item.find(f".//*[@id='{url.fragment}']")
                else:
                    return item.find(url.fragment)
            else:
                return item
        elif fmt in [".json", ".ipynb"]:
            logger.info(f"Resolving URI: {ref}")
            with open(url.path,"r") as f:
                item = json.load(f)
            if url.fragment:
                return resolve_fragment(item, url.fragment, "jsonpointer")
            else:
                return item
        elif fmt in ["yaml",".yaml", "yml", ".yml"]:
            with open(url.path,"r") as f:
                item = yaml.load(f, Loader=yaml.Loader)
            if url.fragment:
                return resolve_fragment(item, url.fragment, "jsonpointer")
            else:
                return item
        elif fmt in [".rst"]:
            document: ElementTree = proc_rst.rst_to_xml(url.path)
            if url.fragment:
                return resolve_fragment(document, url.fragment, "xpath")
            else:
                return document
    else:
        raise FileNotFoundError(url.path)

def resolve_fragment(item, fragment, scheme):
    # if fragment[0] in ["%","@"]:
    #     return resolve_fpointer(item, fragment)
    if scheme.lower() in ["jsonpointer"]:
        return aurore.jsonpointer.resolve_pointer(item, fragment)
        # for key in fragment.split("/"):
        #     if key: item = item[key]
    elif scheme.lower() == "xpath":
        if "/@" in fragment:
            fragment, attrib = fragment.split("/@")
        else:
            attrib=None
        elem = item.find(fragment)

        if elem is not None:
            if attrib:
                return elem.attrib[attrib]
            else:
                return elem.text 
        else:
            return elem

    return item

def resolve_xpath(item, fragid):
    pass 




import io, os, re
from glob import glob
from os.path import splitext, basename, join, dirname

from setuptools import Extension
from setuptools import find_packages
from setuptools import setup









def get_version(package:str)->str:
    """Return package version as listed in `__version__` in `__init__.py`."""
    with open(os.path.join(package, '__init__.py')) as f:
        init_py = f.read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


def read(*names, **kwargs):
    with io.open(
        join(dirname(__file__), *names),
        encoding=kwargs.get('encoding', 'utf8')
    ) as fh:
        return fh.read()

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="aurore",
    version=get_version("./src/aurore/"),
    author="Claudio Perez",
    author_email="claudio_perez@berkeley.edu",
    description="Utilities",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/claudioperez/aurore",
    py_modules=[splitext(basename(path))[0] for path in glob("src/*.py")],
    include_package_data=True,
    packages=find_packages('src'),
    package_dir={'': 'src'},
    entry_points={
        'console_scripts': [
            'aurore = aurore.__main__:main',
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "docutils",
        "coloredlogs",
        "pyyaml",
        "jsonpath2"
        #"rendre"
    ]
)



"""
Copyright (c) 2015, Waylan Limberg
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or other
materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its contributors may
be used to endorse or promote products derived from this software without
specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
OF THE POSSIBILITY OF SUCH DAMAGE.

MultiMarkdown Meta-Data

Extracts, parses and transforms MultiMarkdown style data from documents.

"""


import re
import yaml
try:
    from yaml import CSafeLoader as SafeLoader
except ImportError:  # pragma: no cover
    from yaml import SafeLoader

#####################################################################
# Data Parser                                                       #
#####################################################################

YAML_RE = re.compile(r'^-{3}[ \t]*\n(.*?\n)(?:\.{3}|-{3})[ \t]*\n', re.UNICODE | re.DOTALL)
META_RE = re.compile(r'^[ ]{0,3}(?P<key>[A-Za-z0-9_-]+):\s*(?P<value>.*)')
META_MORE_RE = re.compile(r'^([ ]{4}|\t)(\s*)(?P<value>.*)')


def get_data(doc,include_dirs=None):
    """
    Extract meta-data from a text document.

    Returns a tuple of document and a data dict.
    """
    data = {}

    # First try YAML
    m = YAML_RE.match(doc)
    if m:
        try:
            import yeux
            data = yeux.load(m.group(1),include="ccorp",include_dirs=include_dirs)
            if isinstance(data, dict):
                doc = doc[m.end():].lstrip('\n')
            else:
                data = {}
        except Exception:
            data = yaml.load(m.group(1), SafeLoader)
            if isinstance(data, dict):
                doc = doc[m.end():].lstrip('\n')
            else:
                data = {}
        except Exception:
            pass
        return doc, data

    # No YAML deliminators. Try MultiMarkdown style
    lines = doc.replace('\r\n', '\n').replace('\r', '\n').split('\n')

    key = None
    while lines:
        line = lines.pop(0)

        if line.strip() == '':
            break  # blank line - done
        m1 = META_RE.match(line)
        if m1:
            key = m1.group('key').lower().strip()
            value = m1.group('value').strip()
            if key in data:
                data[key] += ' {}'.format(value)
            else:
                data[key] = value
        else:
            m2 = META_MORE_RE.match(line)
            if m2 and key:
                # Add another line to existing key
                data[key] += ' {}'.format(m2.group('value').strip())
            else:
                lines.insert(0, line)
                break  # no meta data - done
    return '\n'.join(lines).lstrip('\n'), data



def _copy_tree(entries, src, dst, symlinks, ignore, copy_function,
              ignore_dangling_symlinks, dirs_exist_ok=False):
    if ignore is not None:
        ignored_names = ignore(os.fspath(src), [x.name for x in entries])
    else:
        ignored_names = set()

    os.makedirs(dst, exist_ok=dirs_exist_ok)
    errors = []
    use_srcentry = copy_function is copy2 or copy_function is copy

    for srcentry in entries:
        if srcentry.name in ignored_names:
            continue
        srcname = os.path.join(src, srcentry.name)
        dstname = os.path.join(dst, srcentry.name)
        srcobj = srcentry if use_srcentry else srcname
        try:
            is_symlink = srcentry.is_symlink()
            if is_symlink and os.name == 'nt':
                # Special check for directory junctions, which appear as
                # symlinks but we want to recurse.
                lstat = srcentry.stat(follow_symlinks=False)
                if lstat.st_reparse_tag == stat.IO_REPARSE_TAG_MOUNT_POINT:
                    is_symlink = False
            if is_symlink:
                pass
                # linkto = os.readlink(srcname)
                # if symlinks:
                #     # We can't just leave it to `copy_function` because legacy
                #     # code with a custom `copy_function` may rely on copy_tree
                #     # doing the right thing.
                #     os.symlink(linkto, dstname)
                #     copystat(srcobj, dstname, follow_symlinks=not symlinks)
                # else:
                #     # ignore dangling symlink if the flag is on
                #     if not os.path.exists(linkto) and ignore_dangling_symlinks:
                #         continue
                #     # otherwise let the copy occur. copy2 will raise an error
                #     if srcentry.is_dir():
                #         copy_tree(srcobj, dstname, symlinks, ignore,
                #                  copy_function, dirs_exist_ok=dirs_exist_ok)
                #     else:
                #         copy_function(srcobj, dstname)
            elif srcentry.is_dir():
                copy_tree(srcobj, dstname, symlinks, ignore, copy_function,
                         dirs_exist_ok=dirs_exist_ok)
            else:
                # Will raise a SpecialFileError for unsupported file types
                copy_function(srcobj, dstname)
        # catch the Error from the recursive copy_tree so that we can
        # continue with other files
        except Error as err:
            errors.extend(err.args[0])
        except OSError as why:
            errors.append((srcname, dstname, str(why)))
    try:
        copystat(src, dst)
    except OSError as why:
        # Copying file access times may fail on Windows
        if getattr(why, 'winerror', None) is None:
            errors.append((src, dst, str(why)))
    if errors:
        raise Error(errors)
    return dst

def copy_tree(src, dst, symlinks=False, ignore=None, copy_function=copy_file,
             ignore_dangling_symlinks=False, dirs_exist_ok=True):
    """Recursively copy a directory tree and return the destination directory.
    dirs_exist_ok dictates whether to raise an exception in case dst or any
    missing parent directory already exists.
    If exception(s) occur, an Error is raised with a list of reasons.
    If the optional symlinks flag is true, symbolic links in the
    source tree result in symbolic links in the destination tree; if
    it is false, the contents of the files pointed to by symbolic
    links are copied. If the file pointed by the symlink doesn't
    exist, an exception will be added in the list of errors raised in
    an Error exception at the end of the copy process.
    You can set the optional ignore_dangling_symlinks flag to true if you
    want to silence this exception. Notice that this has no effect on
    platforms that don't support os.symlink.
    The optional ignore argument is a callable. If given, it
    is called with the `src` parameter, which is the directory
    being visited by copy_tree(), and `names` which is the list of
    `src` contents, as returned by os.listdir():
        callable(src, names) -> ignored_names
    Since copy_tree() is called recursively, the callable will be
    called once for each directory that is copied. It returns a
    list of names relative to the `src` directory that should
    not be copied.
    The optional copy_function argument is a callable that will be used
    to copy each file. It will be called with the source path and the
    destination path as arguments. By default, copy2() is used, but any
    function that supports the same signature (like copy()) can be used.
    """
    sys.audit("shutil.copytree", src, dst)
    with os.scandir(src) as itr:
        entries = list(itr)
    return shutil._copy_tree(entries=entries, src=src, dst=dst, symlinks=symlinks,
                     ignore=ignore, copy_function=copy_function,
                     ignore_dangling_symlinks=ignore_dangling_symlinks,
                     dirs_exist_ok=dirs_exist_ok)



#/usr/bin/env python


import os
import re
import sys
import argparse
import distutils, shutil, logging
from distutils import dir_util
import xml.etree.ElementTree as ElementTree
import urllib.parse
from pathlib import Path

import yaml

from .api import \
    post_init, post_close, \
    get_init, get_item, get_close

from .uri_utils import resolve_uri
from .utils import norm_join
from aurore import config


from aurore import __version__ #= "0.0.1"

def pass_item(item, args, config, accum:dict)->dict:
    return accum

def _apply_filter(rsrc:dict,filter:dict)->set:
    if "match" in filter:
        for key, pattern in filter["match"].items():
            if re.match(pattern,rsrc[key]):
                continue
            else:
                return False
    return set(filter["rules"]) if "rules" in filter else True

def apply_field_filters(resource:ElementTree,filters:dict)->bool:
    matches = []
    for field,fltrs in filters.items():
        j_matches = []
        if ":" in field:
            elem_name, attrib = field.split(":")
            for fltr in fltrs:
                fltr_j_matched = False
                for el in resource.findall(elem_name):
                    if fltr.search(el.attrib[attrib]):
                        fltr_j_matched = True
                        break
                j_matches.append(fltr_j_matched)
        matches.append(all(j_matches))
    return all(matches)


def main(args:list=None)->int:
    parser = argparse.ArgumentParser(prog='aurore', description="Distributed document-oriented DBMS.")
    parser.add_argument("-D","--database-file",
            nargs="?",
            default=".aurore/aurore.db.xml",
            help="Specify a database file"
    )
    parser.add_argument("-C","--config-file",
            default=".aurore/config.yml",
            help="Specify a config file"
    )
    # parser.add_argument("-I","--include-item", nargs="?", action="append")
    # parser.add_argument("-E","--exclude",nargs="*", action="extend")
    parser.add_argument("-B","--base-uri", default="")
    parser.add_argument("-d","--defaults",
        help="Specify a default yaml file.",
        nargs="?",
        action="append"
    )
    parser.add_argument("-v","--verbose",
        help="Generate verbose logging output.",
        action="count",
        default=0
    )
    parser.add_argument("-q","--quiet", action="store_true")
    parser.add_argument("--version",
        help="Output version information and exit.",
        action="store_true")

    subparsers = parser.add_subparsers(title='subcommands') #,description='list of subcommands',help='additional help')

    #-Post-------------------------------------------------------------
    post_parser = subparsers.add_parser("post", help="Create new entity [INCOMPLETE]")
    post_parser.add_argument("type")
    post_parser.add_argument("location", nargs="?", default="$$lid/metadata.xml")
    post_parser.set_defaults(initfunc=post_init)
    post_parser.set_defaults(func=None)
    post_parser.set_defaults(closefunc=post_close)

    #-Get-------------------------------------------------------------
    get_parser = subparsers.add_parser("get", help="Retrieve data from entities")
    get_parser.add_argument("-p","--rel-path")
    get_parser.add_argument("item-selectors", nargs="*", default=["*"])
    get_parser.set_defaults(initfunc=get_init)
    get_parser.set_defaults(func=get_item)
    get_parser.set_defaults(closefunc=get_close)


    #-----------------------------------------------------------------
    # Main
    #-----------------------------------------------------------------
    if args is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(args)

    #-Logging-----------------------------------------------------------
    logger = logging.getLogger("aurore")
    levels = [logging.ERROR, logging.WARN, logging.INFO, logging.DEBUG]
    levelNames = ["ERROR", "WARN", "INFO", "DEBUG"]
    logger.setLevel(levels[
        0 if args.quiet else min(args.verbose,3) if args.verbose else 0
        ])
    try:
        import coloredlogs
        coloredlogs.install(level=levelNames[
            0 if args.quiet else min(args.verbose,3) if args.verbose else 0
            ])
    except:
        pass

    if args.version:
        print(__version__)
        exit(0)

    #-Config----------------------------------------------------------
    cfg = config.load_config(args.config_file)

    #-Main-----------------------------------------------------------

    accum = {}
    if args.base_uri:
        pass
    else:
        args.base_uri = cfg["base_uri"]

    logger.info(f"{args}")

    if args.database_file =="-":
        # logger.debug(f"STDIN:\n{sys.stdin.read()}")
        # tree = ElementTree.ElementTree(ElementTree.fromstring(sys.stdin.read()))
        tree = ElementTree.parse(sys.stdin)
    else:
        tree = ElementTree.parse(args.database_file)
    root = tree.getroot()
    items = root.find("items")
    if items is None:
        items = root
    if not args.base_uri and "base" in items.attrib:
        args.base_uri = root.find("items").attrib["base"]

    category_schemes = tree.findall(".//category-scheme")
    if "category_file" in args and args.category_file:
        categories = ElementTree.parse(args.category_file[0])
        category_schemes.append(categories.findall("category-scheme"))


    setattr(args, "category_schemes", category_schemes)
    logger.debug(f"category schemes: {category_schemes}")

    tag_schemes = root.find("tag-scheme")
    setattr(args, "tag_schemes", tag_schemes)

    FILTERS = {}

    #-------------------------------------------------------------------
    if "initfunc" in args:
        accum = args.initfunc(args,cfg)
    else:
        accum = {}

    if args.func:
        for item in items:
            if "src" in item.attrib:
                resource_path = norm_join(args.base_uri, item.attrib["src"])
                logger.info(f"Resource path: {resource_path}")
                resource = resolve_uri(resource_path)
                resource.attrib.update({"id": item.attrib["id"]})
            
            elif "type" in item.attrib:
                xml_string = cfg["types"][item.attrib["type"]].render(**item.attrib)
                resource = ElementTree.fromstring(xml_string)
            
            else:
                resource = item

            if apply_field_filters(resource,FILTERS):
                logger.info("Entering {}".format(item.attrib["id"]))
                logger.debug(f"Item: {item}")
                key = ElementTree.Element("str")
                key.attrib.update({"key": "id"})
                key.text = item.attrib["id"]
                resource.insert(0, key)
                accum = args.func(resource, args, cfg, accum)


    if "closefunc" in args:
        args.closefunc(args, cfg, accum)
    
    return 0

if __name__ == "__main__": main()
